"""`coordinax.vecs` Module."""

__all__ = [
    # API
    "vector",
    "vconvert",
    "normalize_vector",
    "cartesian_vector_type",
    "time_derivative_vector_type",
    "time_antiderivative_vector_type",
    "time_nth_derivative_vector_type",
    "IrreversibleDimensionChange",
    # Base
    "AbstractVector",
    "AttrFilter",
    "VectorAttribute",
    "ToUnitsOptions",
    # Base Classes
    "AbstractPos",
    "AbstractVel",
    "AbstractAcc",
    # --- 1D ---
    # Base
    "AbstractPos1D",
    "AbstractVel1D",
    "AbstractAcc1D",
    # Radial
    "RadialPos",
    "RadialVel",
    "RadialAcc",
    # Cartesian
    "CartesianPos1D",
    "CartesianVel1D",
    "CartesianAcc1D",
    # --- 2D ---
    # Base
    "AbstractPos2D",
    "AbstractVel2D",
    "AbstractAcc2D",
    # TwoSphere
    "TwoSpherePos",
    "TwoSphereVel",
    "TwoSphereAcc",
    # Polar
    "PolarPos",
    "PolarVel",
    "PolarAcc",
    # Cartesian
    "CartesianPos2D",
    "CartesianVel2D",
    "CartesianAcc2D",
    # --- 3D ---
    # Base
    "AbstractPos3D",
    "AbstractVel3D",
    "AbstractAcc3D",
    # Cartesian
    "CartesianPos3D",
    "CartesianVel3D",
    "CartesianAcc3D",
    # Cylindrical
    "CylindricalPos",
    "CylindricalVel",
    "CylindricalAcc",
    # Base Spherical
    "AbstractSphericalPos",
    "AbstractSphericalVel",
    "AbstractSphericalAcc",
    # Spherical
    "SphericalPos",
    "SphericalVel",
    "SphericalAcc",
    # Math Spherical
    "MathSphericalPos",
    "MathSphericalVel",
    "MathSphericalAcc",
    # LonLat Spherical
    "LonLatSphericalPos",
    "LonLatSphericalVel",
    "LonLatSphericalAcc",
    "LonCosLatSphericalVel",
    # Prolate Spheroidal
    "ProlateSpheroidalPos",
    "ProlateSpheroidalVel",
    "ProlateSpheroidalAcc",
    # Generic
    "Cartesian3D",
    # --- 4D ---
    "AbstractPos4D",
    "FourVector",
    # --- N-D ---
    "AbstractPosND",
    "AbstractVelND",
    "AbstractAccND",
    "CartesianPosND",
    "CartesianVelND",
    "CartesianAccND",
    "PoincarePolarVector",
    # --- Space ---
    "AbstractVectors",
    "Space",
    # --- Misc ---
    "POSITION_CLASSES",
    "VELOCITY_CLASSES",
    "ACCELERATION_CLASSES",
]

from jaxtyping import install_import_hook

from .setup_package import RUNTIME_TYPECHECKER

with install_import_hook("coordinax.vecs", RUNTIME_TYPECHECKER):
    from ._src import vectors
    from ._src.vectors.api import (
        cartesian_vector_type,
        normalize_vector,
        time_antiderivative_vector_type,
        time_derivative_vector_type,
        time_nth_derivative_vector_type,
        vconvert,
        vector,
    )
    from ._src.vectors.base import (
        AbstractVector,
        AttrFilter,
        ToUnitsOptions,
        VectorAttribute,
    )
    from ._src.vectors.base_acc import ACCELERATION_CLASSES, AbstractAcc
    from ._src.vectors.base_pos import POSITION_CLASSES, AbstractPos
    from ._src.vectors.base_vel import VELOCITY_CLASSES, AbstractVel
    from ._src.vectors.d1 import (
        AbstractAcc1D,
        AbstractPos1D,
        AbstractVel1D,
        CartesianAcc1D,
        CartesianPos1D,
        CartesianVel1D,
        RadialAcc,
        RadialPos,
        RadialVel,
    )
    from ._src.vectors.d2 import (
        AbstractAcc2D,
        AbstractPos2D,
        AbstractVel2D,
        CartesianAcc2D,
        CartesianPos2D,
        CartesianVel2D,
        PolarAcc,
        PolarPos,
        PolarVel,
        TwoSphereAcc,
        TwoSpherePos,
        TwoSphereVel,
    )
    from ._src.vectors.d3 import (
        AbstractAcc3D,
        AbstractPos3D,
        AbstractSphericalAcc,
        AbstractSphericalPos,
        AbstractSphericalVel,
        AbstractVel3D,
        Cartesian3D,
        CartesianAcc3D,
        CartesianPos3D,
        CartesianVel3D,
        CylindricalAcc,
        CylindricalPos,
        CylindricalVel,
        LonCosLatSphericalVel,
        LonLatSphericalAcc,
        LonLatSphericalPos,
        LonLatSphericalVel,
        MathSphericalAcc,
        MathSphericalPos,
        MathSphericalVel,
        ProlateSpheroidalAcc,
        ProlateSpheroidalPos,
        ProlateSpheroidalVel,
        SphericalAcc,
        SphericalPos,
        SphericalVel,
    )
    from ._src.vectors.d4 import AbstractPos4D, FourVector
    from ._src.vectors.dn import (
        AbstractAccND,
        AbstractPosND,
        AbstractVelND,
        CartesianAccND,
        CartesianPosND,
        CartesianVelND,
        PoincarePolarVector,
    )
    from ._src.vectors.exceptions import IrreversibleDimensionChange
    from ._src.vectors.space import AbstractVectors, Space


del vectors, install_import_hook, RUNTIME_TYPECHECKER



"""Transformations between representations."""

__all__: list[str] = []

from typing import Any, cast

import equinox as eqx
import jax
import jax.tree as jtu
from jaxtyping import Array, Float, Real
from plum import dispatch

import quaxed.numpy as jnp
import unxt as u
from unxt.quantity import is_any_quantity

import coordinax._src.vectors.custom_types as ct
from coordinax._src.distances import AbstractDistance
from coordinax._src.vectors import api, d1, d2, d3
from coordinax._src.vectors.base_acc import AbstractAcc
from coordinax._src.vectors.base_pos import AbstractPos
from coordinax._src.vectors.base_vel import AbstractVel

# NOTE: this is using Quantities. Using raw arrays is ~20x faster.
# TODO: what about the auxiliary values?
pos_jac_fn = jax.vmap(
    jax.jacfwd(api.vconvert, argnums=2, has_aux=True), in_axes=(None, None, 0)
)

is_q_or_arr = lambda x: is_any_quantity(x) or eqx.is_array(x)  # noqa: E731


def transform_jac(
    jac: dict[str, u.AbstractQuantity],
) -> dict[str, dict[str, u.AbstractQuantity]]:
    """Transform Jacobian.

    It comes in as  ``{k: Quantity({kk: Quantity(vv, unit2)}, unit1)}``
    it needs to be rearranged to ``{k: {kk: Quantity(vv, unit1/unit2)}}``.

    """
    return {
        out_k: {
            k: u.Quantity(v.value, out_v.unit / v.unit) for k, v in out_v.value.items()
        }
        for out_k, out_v in jac.items()
    }


@jax.jit
def dot_jac_vec(
    jac: dict[str, dict[str, u.AbstractQuantity]], vec: dict[str, u.AbstractQuantity]
) -> dict[str, u.AbstractQuantity]:
    """Dot product of a Jacobian dict and a vector dict.

    This is a helper function for the `vconvert` function.

    Examples
    --------
    >>> import unxt as u

    >>> J = {"r": {"x": u.Quantity(1, ""), "y": u.Quantity(2, "")},
    ...      "phi": {"x": u.Quantity(3, "rad/km"), "y": u.Quantity(4, "rad/km")}}

    >>> v = {"x": u.Quantity(1, "km"), "y": u.Quantity(2, "km")}

    >>> dot_jac_vec(J, v)
    {'phi': Quantity['angle'](Array(11, dtype=int32, ...), unit='rad'),
     'r': Quantity['length'](Array(5, dtype=int32, ...), unit='km')}

    """
    flat_jac, treedef = eqx.tree_flatten_one_level(jac)
    flat_dotted = jtu.map(
        lambda inner: jtu.reduce(
            jnp.add,
            jtu.map(jnp.multiply, inner, vec, is_leaf=is_q_or_arr),
            is_leaf=is_q_or_arr,
        ),
        flat_jac,
        is_leaf=lambda v: isinstance(v, dict),
    )
    return jtu.unflatten(treedef, flat_dotted)


def atleast_1d_float(x: Real[Array, "..."]) -> Float[Array, "..."]:
    """Convert to 1D float array."""
    return jnp.atleast_1d(jnp.astype(x, float, copy=False))


# ============================================================================


@dispatch.multi(
    (type[AbstractVel], type[AbstractVel], ct.ParamsDict, ct.ParamsDict),
)
def vconvert(
    to_vel_cls: type[AbstractVel],
    from_vel_cls: type[AbstractVel],
    p_vel: ct.ParamsDict,
    p_pos: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
) -> tuple[ct.ParamsDict, ct.OptAuxDict]:
    """AbstractVel1D -> AbstractVel1D.

    Parameters
    ----------
    to_vel_cls
        The target type of the vector differential.
    from_vel_cls
        The type of the vector differential to transform.
    p_vel
        The data of the vector differential to transform.
    p_pos
        The data of the position vector used to transform the differential.
    in_aux
        The auxiliary data of the vector differential to transform.
    out_aux
        The auxiliary data of the position vector used to transform the
        differential.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    Let's start in 1D:

    >>> q = {"x": u.Quantity([1.0], "km")}
    >>> p = {"x": u.Quantity([1.0], "km/s")}
    >>> newp = cxv.vconvert(cxv.RadialVel, cxv.CartesianVel1D, p, q)
    >>> print(newp)
    ({'r': Quantity['speed'](Array([1.], dtype=float32), unit='km / s')}, {})

    """
    # Check the dimensionality
    to_vel_cls = eqx.error_if(
        to_vel_cls,
        from_vel_cls._dimensionality() != to_vel_cls._dimensionality(),  # noqa: SLF001
        "Dimensionality mismatch",
    )

    shape = jnp.broadcast_shapes(*[v.shape for v in p_vel.values()])

    # The position is assumed to be in the type required by the differential to
    # construct the Jacobian. E.g. for CartesianVel1D -> RadialVel, we need the
    # Jacobian of the CartesianPos1D -> RadialPos transform.
    p_pos = jtu.map(atleast_1d_float, p_pos)
    p_pos = {  # NOTE: if use unitful jacobian, this is not needed
        k: (v.distance if isinstance(v, AbstractDistance) else v)
        for k, v in p_pos.items()
    }

    # -----------------------
    # Compute the Jacobian of the position transformation.
    to_pos_cls = to_vel_cls.time_antiderivative_cls
    from_pos_cls = from_vel_cls.time_antiderivative_cls
    jac, _ = pos_jac_fn(to_pos_cls, from_pos_cls, p_pos, in_aux=in_aux)
    jac = transform_jac(jac)

    # -----------------------
    # Transform the velocity

    to_p_vel = dot_jac_vec(jac, p_vel)
    to_p_vel = jtu.map(lambda x: jnp.reshape(x, shape), to_p_vel)

    return to_p_vel, (out_aux or {})


# TODO: implement for cross-representations
@dispatch.multi(
    # N-D -> N-D
    (type[d1.AbstractVel1D], d1.AbstractVel1D, AbstractPos | u.AbstractQuantity),
    (type[d2.AbstractVel2D], d2.AbstractVel2D, AbstractPos | u.AbstractQuantity),
    (type[d3.AbstractVel3D], d3.AbstractVel3D, AbstractPos | u.AbstractQuantity),
)
def vconvert(
    to_vel_cls: type[AbstractVel],
    from_vel: AbstractVel,
    from_pos: AbstractPos | u.AbstractQuantity,
    /,
    **kwargs: Any,
) -> AbstractVel:
    """AbstractVel -> Cartesian -> AbstractVel.

    This is the base case for the transformation of vector differentials.

    Parameters
    ----------
    to_vel_cls
        The target type of the vector differential.
    from_vel
        The vector differential to transform.
    from_pos
        The position vector used to transform the differential.
    **kwargs
        Additional keyword arguments.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    Let's start in 1D:

    >>> q = cxv.CartesianPos1D.from_(1.0, "km")
    >>> p = cxv.CartesianVel1D.from_(1.0, "km/s")
    >>> newp = cxv.vconvert(cxv.RadialVel, p, q)
    >>> print(newp)
    <RadialVel (r[km / s])
        [1.]>

    Now in 2D:

    >>> q = cxv.CartesianPos2D.from_([1.0, 2.0], "km")
    >>> p = cxv.CartesianVel2D.from_([1.0, 2.0], "km/s")
    >>> newp = cxv.vconvert(cxv.PolarVel, p, q)
    >>> print(newp)
    <PolarVel (r[km / s], phi[rad / s])
        [2.236 0.   ]>

    And in 3D:

    >>> q = cxv.CartesianPos3D.from_([1.0, 2.0, 3.0], "km")
    >>> p = cxv.CartesianVel3D.from_([1.0, 2.0, 3.0], "km/s")
    >>> newp = cxv.vconvert(cxv.SphericalVel, p, q)
    >>> print(newp)
    <SphericalVel (r[km / s], theta[rad / s], phi[rad / s])
        [ 3.742e+00 -8.941e-08  0.000e+00]>

    If given a position as a Quantity, it will be converted to the appropriate
    Cartesian vector:

    >>> p = cxv.CartesianVel3D.from_([1.0, 2.0, 3.0], "km/s")
    >>> newp = cxv.vconvert(cxv.SphericalVel, p, u.Quantity([1.0, 2.0, 3.0], "km"))
    >>> print(newp)
    <SphericalVel (r[km / s], theta[rad / s], phi[rad / s])
        [ 3.742e+00 -8.941e-08  0.000e+00]>

    """
    shape = from_vel.shape

    # ----------------------------
    # Prepare the position

    in_pos_cls = from_vel.time_antiderivative_cls

    # Parse the position to an AbstractPos
    from_posv_ = cast(
        AbstractPos,
        from_pos
        if isinstance(from_pos, AbstractPos)
        else in_pos_cls.cartesian_type.from_(from_pos),
    )

    # Transform the position to the type required by the differential to
    # construct the Jacobian. E.g. if we are transforming CartesianVel1D ->
    # RadialVel, we need the Jacobian of the CartesianPos1D -> RadialPos so the
    # position must be transformed to CartesianPos1D.
    from_posv = vconvert(in_pos_cls, from_posv_, **kwargs)

    # ----------------------------
    # Perform the transformation

    # TODO: add  df(q)/dt, which is 0 for all current transforms
    to_vel_params, to_vel_aux = vconvert(
        to_vel_cls,
        type(from_vel),
        from_vel.asdict(),
        from_posv.asdict(),
    )

    # ----------------------------
    # Reconstruct the vector

    to_vel = to_vel_cls(**to_vel_params, **to_vel_aux)
    to_vel = to_vel.reshape(shape)  # reshape to original shape

    return to_vel  # noqa: RET504


# ===================================================================


@dispatch
def vconvert(
    to_acc_cls: type[AbstractAcc],
    from_acc_cls: type[AbstractAcc],
    p_acc: ct.ParamsDict,
    p_vel: ct.ParamsDict,
    p_pos: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
) -> tuple[ct.ParamsDict, ct.OptAuxDict]:
    """AbstractAcc1D -> AbstractAcc1D.

    Parameters
    ----------
    to_acc_cls
        The target type of the vector differential.
    from_acc_cls
        The type of the vector differential to transform.
    p_acc
        The data of the vector differential to transform.
    p_vel, p_pos
        The data of the velocity/position vector used to transform the
        differential.
    in_aux
        The auxiliary data of the vector differential to transform.
    out_aux
        The auxiliary data of the position vector used to transform the
        differential.
    units
        The unit system to use for the transformation.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    Let's start in 1D:

    >>> q = {"x": u.Quantity([1.0], "km")}
    >>> p = {"x": u.Quantity([1.0], "km/s")}
    >>> a = {"x": u.Quantity([1.0], "km/s2")}
    >>> newa = cxv.vconvert(cxv.RadialAcc, cxv.CartesianAcc1D, a, p, q)
    >>> print(newa)
    ({'r': Quantity['acceleration'](Array([1.], dtype=float32), unit='km / s2')}, {})

    """
    # Check the dimensionality
    to_acc_cls = eqx.error_if(
        to_acc_cls,
        from_acc_cls._dimensionality() != to_acc_cls._dimensionality(),  # noqa: SLF001
        "Dimensionality mismatch",
    )

    shape = jnp.broadcast_shapes(*[v.shape for v in p_acc.values()])

    # The position is assumed to be in the type required by the differential to
    # construct the Jacobian. E.g. for CartesianVel1D -> RadialVel, we need the
    # Jacobian of the CartesianPos1D -> RadialPos transform.
    p_pos = jtu.map(atleast_1d_float, p_pos)
    p_pos = {  # NOTE: if use unitful jacobian, this is not needed
        k: (v.distance if isinstance(v, AbstractDistance) else v)
        for k, v in p_pos.items()
    }

    # -----------------------
    # Compute the Jacobian of the position transformation.
    to_pos_cls = to_acc_cls.time_nth_derivative_cls(-2)
    from_pos_cls = from_acc_cls.time_nth_derivative_cls(-2)
    jac, _ = pos_jac_fn(to_pos_cls, from_pos_cls, p_pos, in_aux=in_aux)
    jac = transform_jac(jac)

    # -----------------------
    # Transform the acceleration

    to_p_acc = dot_jac_vec(jac, p_acc)
    to_p_acc = jtu.map(lambda x: jnp.reshape(x, shape), to_p_acc)

    return to_p_acc, (out_aux or {})


# TODO: implement for cross-representations
@dispatch.multi(
    # N-D -> N-D
    (
        type[d1.AbstractAcc1D],
        d1.AbstractAcc1D,
        AbstractVel | u.AbstractQuantity,
        AbstractPos | u.AbstractQuantity,
    ),
    (
        type[d2.AbstractAcc2D],
        d2.AbstractAcc2D,
        AbstractVel | u.AbstractQuantity,
        AbstractPos | u.AbstractQuantity,
    ),
    (
        type[d3.AbstractAcc3D],
        d3.AbstractAcc3D,
        AbstractVel | u.AbstractQuantity,
        AbstractPos | u.AbstractQuantity,
    ),
)
def vconvert(
    to_acc_cls: type[AbstractAcc],
    from_acc: AbstractAcc,
    from_vel: AbstractVel | u.Quantity["speed"],
    from_pos: AbstractPos | u.Quantity["length"],
    /,
    **kwargs: Any,
) -> AbstractAcc:
    r"""AbstractAcc -> Cartesian -> AbstractAcc.

    This is the base case for the transformation of accelerations.

    Let $\mathbf{x}$ be a position vector in one representation $\mathbf{y}$ a
    position vector in another representation related by:

    $$ \mathbf{y} = f(\mathbf{x}) $$

    where $f$ is a differentiable function mapping between the coordinate
    systems.

    The Jacobian matrix $J$ of the transformation is:

    $$ J = \frac{\partial \mathbf{y}}{\partial \mathbf{x}}$$

    The coordinate transformation of the acceleration is given by the chain
    rule:

    $$ \ddot{\mathbf{y}} = \dot{J} \dot{\mathbf{x}} + J \ddot{\mathbf{x}}$$

    where $\dot{J}$ is the time derivative of the Jacobian matrix. This function
    assumes that the representation conversion is time-invariant, so $\dot{J} =
    0$. Thus, the transformation simplifies to:

    $$ \ddot{\mathbf{y}} = J \ddot{\mathbf{x}}$$

    This function implements this transformation.

    Parameters
    ----------
    to_acc_cls
        The target type of the vector acceleration.
    from_acc
        The vector acceleration to transform.
    from_vel
        The velocity vector used to transform the acceleration.
    from_pos
        The position vector used to transform the acceleration.
    **kwargs
        Additional keyword arguments.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    Let's start in 1D:

    >>> q = cx.vecs.CartesianPos1D(x=u.Quantity(1.0, "km"))
    >>> p = cx.vecs.CartesianVel1D(x=u.Quantity(1.0, "km/s"))
    >>> a = cx.vecs.CartesianAcc1D(x=u.Quantity(1.0, "km/s2"))
    >>> cx.vconvert(cx.vecs.RadialAcc, a, p, q)
    RadialAcc( r=Quantity[...](value=f32[], unit=Unit("km / s2")) )

    Now in 2D:

    >>> q = cx.vecs.CartesianPos2D.from_([1.0, 2.0], "km")
    >>> p = cx.vecs.CartesianVel2D.from_([1.0, 2.0], "km/s")
    >>> a = cx.vecs.CartesianAcc2D.from_([1.0, 2.0], "km/s2")
    >>> cx.vconvert(cx.vecs.PolarAcc, a, p, q)
    PolarAcc(
      r=Quantity[...](value=f32[], unit=Unit("km / s2")),
      phi=Quantity[...]( value=f32[], unit=Unit("rad / s2") )
    )

    And in 3D:

    >>> q = cx.CartesianPos3D.from_([1.0, 2.0, 3.0], "km")
    >>> p = cx.CartesianVel3D.from_([1.0, 2.0, 3.0], "km/s")
    >>> a = cx.vecs.CartesianAcc3D.from_([1.0, 2.0, 3.0], "km/s2")
    >>> cx.vconvert(cx.vecs.SphericalAcc, a, p, q)
    SphericalAcc(
      r=Quantity[...](value=f32[], unit=Unit("km / s2")),
      theta=Quantity[...]( value=f32[], unit=Unit("rad / s2") ),
      phi=Quantity[...]( value=f32[], unit=Unit("rad / s2") )
    )

    If given a position as a Quantity, it will be converted to the appropriate
    Cartesian vector:

    >>> cx.vconvert(cx.vecs.SphericalAcc, a,
    ...             u.Quantity([1.0, 2.0, 3.0], "km/s"),
    ...             u.Quantity([1.0, 2.0, 3.0], "km"))
    SphericalAcc(
      r=Quantity[...](value=f32[], unit=Unit("km / s2")),
      theta=Quantity[...]( value=f32[], unit=Unit("rad / s2") ),
      phi=Quantity[...]( value=f32[], unit=Unit("rad / s2") )
    )

    """
    shape = from_acc.shape

    # ----------------------------
    # Prepare the position

    # Parse the position to an AbstractPos
    in_pos_cls = from_acc.time_nth_derivative_cls(-2)
    from_posv_ = cast(
        AbstractPos,
        from_pos
        if isinstance(from_pos, AbstractPos)
        else in_pos_cls.cartesian_type.from_(from_pos),
    )

    # Transform the position to the type required by the differential to
    # construct the Jacobian. E.g. if we are transforming CartesianVel1D ->
    # RadialVel, we need the Jacobian of the CartesianPos1D -> RadialPos so the
    # position must be transformed to CartesianPos1D.
    from_posv = vconvert(in_pos_cls, from_posv_, **kwargs)

    # ----------------------------
    # Prepare the velocity

    in_vel_cls = from_acc.time_antiderivative_cls
    from_velv_ = cast(
        AbstractVel,
        from_vel
        if isinstance(from_vel, AbstractVel)
        else in_vel_cls.cartesian_type.from_(from_vel),
    )
    from_velv = vconvert(in_vel_cls, from_velv_, from_posv, **kwargs)

    # ----------------------------
    # Perform the transformation

    # TODO: add  df(q)/dt, which is 0 for all current transforms
    to_acc_params, to_acc_aux = vconvert(
        to_acc_cls,
        type(from_acc),
        from_acc.asdict(),
        from_velv.asdict(),
        from_posv.asdict(),
    )

    # ----------------------------
    # Reconstruct the vector

    to_acc = to_acc_cls(**to_acc_params, **to_acc_aux)
    to_acc = to_acc.reshape(shape)  # reshape to original shape

    return to_acc  # noqa: RET504



"""Register Distance support for `unxt`."""
# pylint: disable=import-error

__all__: list[str] = []


from plum import dispatch

import unxt as u

from .base import AbstractDistance


@dispatch
def dimension_of(obj: type[AbstractDistance], /) -> u.dims.AbstractDimension:
    """Get the dimension of an angle.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> u.dimension_of(cx.distance.AbstractDistance)
    PhysicalType('length')

    """
    return u.dimension("length")



"""Built-in vector classes."""

__all__ = [
    "AbstractSphericalAcc",
    "AbstractSphericalPos",
    "AbstractSphericalVel",
]


import unxt as u

from .base import AbstractAcc3D, AbstractPos3D, AbstractVel3D

_90d = u.Quantity(90, "deg")
_180d = u.Quantity(180, "deg")
_360d = u.Quantity(360, "deg")


class AbstractSphericalPos(AbstractPos3D):
    """Abstract spherical vector representation."""


class AbstractSphericalVel(AbstractVel3D):
    """Spherical differential representation."""


class AbstractSphericalAcc(AbstractAcc3D):
    """Spherical acceleration representation."""



"""Representation of velocities in different systems."""

__all__ = ["AbstractVel", "VELOCITY_CLASSES"]

import functools as ft
from typing import TYPE_CHECKING, Any, cast

import equinox as eqx

import unxt as u

from coordinax._src.utils import classproperty
from coordinax._src.vectors import api
from coordinax._src.vectors.base import AbstractVector
from coordinax._src.vectors.base_pos import AbstractPos
from coordinax._src.vectors.mixins import AvalMixin

if TYPE_CHECKING:
    import coordinax.vecs


class AbstractVel(AvalMixin, AbstractVector):  # pylint: disable=abstract-method
    """Abstract representation of vector differentials in different systems."""

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Initialize the subclass.

        The subclass is registered.
        """
        VELOCITY_CLASSES_MUTABLE[cls] = None

    # ===============================================================
    # Vector API

    @classproperty
    @classmethod
    def cartesian_type(cls) -> "type[coordinax.vecs.AbstractVel]":
        """Return the corresponding Cartesian vector class."""
        return api.cartesian_vector_type(cls)

    @classproperty
    @classmethod
    def time_derivative_cls(cls) -> "type[coordinax.vecs.AbstractAcc]":
        """Return the corresponding time derivative class."""
        return api.time_derivative_vector_type(cls)

    @classproperty
    @classmethod
    def time_antiderivative_cls(cls) -> type[AbstractPos]:
        """Return the corresponding time antiderivative class."""
        return api.time_antiderivative_vector_type(cls)

    @classmethod
    def time_nth_derivative_cls(
        cls, *, n: int
    ) -> "type[coordinax.vecs.AbstractVector]":
        """Return the corresponding time nth derivative class."""
        return api.time_nth_derivative_vector_type(cls, n=n)

    # ===============================================================
    # Convenience methods

    @ft.partial(eqx.filter_jit)
    def norm(self: "AbstractVel", q: AbstractPos, /) -> u.Quantity["speed"]:
        """Return the norm of the vector.

        Examples
        --------
        >>> import coordinax as cx

        >>> q = cx.vecs.CartesianPos2D.from_([1, 1], "km")
        >>> p = cx.vecs.PolarVel(r=u.Quantity(1, "km/s"), phi=u.Quantity(1, "deg/s"))

        >>> p.norm(q).uconvert('km / s')
        Quantity['speed'](Array(1.0003046, dtype=float32), unit='km / s')

        """
        cart_vel = cast(AbstractVel, self.vconvert(self.cartesian_type, q))
        return cart_vel.norm(q)  # type: ignore[call-arg,misc]


#: Registered velocity classes.
VELOCITY_CLASSES_MUTABLE: dict[type[AbstractVel], None] = {}
VELOCITY_CLASSES = VELOCITY_CLASSES_MUTABLE.keys()



"""Representation of coordinates in different systems."""

__all__: list[str] = []

from typing import Any

import equinox as eqx
import jax
from jaxtyping import ArrayLike
from quax import register

import quaxed.lax as qlax
import quaxed.numpy as jnp
import unxt as u
from dataclassish import replace

from .cartesian import CartesianAcc2D, CartesianPos2D, CartesianVel2D
from .polar import PolarPos
from coordinax._src.vectors.base_pos import AbstractPos

# -----------------------------------------------------


@register(jax.lax.add_p)
def add_p_cart2d_pos(lhs: CartesianPos2D, rhs: AbstractPos, /) -> CartesianPos2D:
    """Add a CartesianPos2D and position vector.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    >>> cart = cx.vecs.CartesianPos2D.from_([1, 2], "km")
    >>> polr = cx.vecs.PolarPos(r=u.Quantity(3, "km"), phi=u.Quantity(90, "deg"))
    >>> print(cart + polr)
    <CartesianPos2D (x[km], y[km])
        [1. 5.]>

    >>> print(jnp.add(cart, polr))
    <CartesianPos2D (x[km], y[km])
        [1. 5.]>

    """
    cart = rhs.vconvert(CartesianPos2D)
    return jax.tree.map(jnp.add, lhs, cart)


@register(jax.lax.add_p)
def add_p_pp(lhs: CartesianVel2D, rhs: CartesianVel2D, /) -> CartesianVel2D:
    """Add two Cartesian velocities.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import coordinax as cx

    >>> v = cx.vecs.CartesianVel2D.from_([1, 2], "km/s")
    >>> print(v + v)
    <CartesianVel2D (x[km / s], y[km / s])
        [2 4]>

    >>> print(jnp.add(v, v))
    <CartesianVel2D (x[km / s], y[km / s])
        [2 4]>

    """
    return jax.tree.map(qlax.add, lhs, rhs)


@register(jax.lax.add_p)
def add_p_aa(lhs: CartesianAcc2D, rhs: CartesianAcc2D, /) -> CartesianAcc2D:
    """Add two Cartesian accelerations.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import coordinax as cx

    >>> v = cx.vecs.CartesianAcc2D.from_([3, 4], "km/s2")
    >>> print(v + v)
    <CartesianAcc2D (x[km / s2], y[km / s2])
        [6 8]>

    >>> print(jnp.add(v, v))
    <CartesianAcc2D (x[km / s2], y[km / s2])
        [6 8]>

    """
    return jax.tree.map(jnp.add, lhs, rhs)


# ------------------------------------------------
# Dot product
# TODO: see implementation in https://github.com/google/tree-math for how to do
# this more generally.


# TODO: this works for more Cartesian types
@register(jax.lax.dot_general_p)
def dot_general_p_cart2d(
    lhs: CartesianPos2D, rhs: CartesianPos2D, /, **kwargs: Any
) -> u.AbstractQuantity:
    """Dot product of two vectors.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    >>> q1 = cx.vecs.CartesianPos2D.from_([1, 2], "m")
    >>> q2 = cx.vecs.CartesianPos2D.from_([3, 4], "m")

    >>> jnp.dot(q1, q2)
    Quantity['area'](Array(11, dtype=int32), unit='m2')

    """
    tree = jax.tree.map(jnp.multiply, lhs, rhs, is_leaf=u.quantity.is_any_quantity)
    return jax.tree.reduce(jnp.add, tree, is_leaf=u.quantity.is_any_quantity)


# ------------------------------------------------


@register(jax.lax.mul_p)
def mul_p_v_cart2d(lhs: ArrayLike, rhs: CartesianPos2D, /) -> CartesianPos2D:
    """Scale a cartesian 2D position by a scalar.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import coordinax as cx

    >>> v = cx.vecs.CartesianPos2D.from_([3, 4], "m")
    >>> jnp.multiply(5, v).x
    Quantity['length'](Array(15, dtype=int32), unit='m')

    """
    # Validation
    lhs = eqx.error_if(
        lhs, any(jax.numpy.shape(lhs)), f"must be a scalar, not {type(lhs)}"
    )

    # Scale the components
    return replace(rhs, x=lhs * rhs.x, y=lhs * rhs.y)


@register(jax.lax.mul_p)
def mul_p_vp(lhs: ArrayLike, rhs: CartesianVel2D, /) -> CartesianVel2D:
    """Scale a cartesian 2D velocity by a scalar.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import coordinax as cx

    >>> v = cx.vecs.CartesianVel2D.from_([3, 4], "m/s")
    >>> print(5 * v)
    <CartesianVel2D (x[m / s], y[m / s])
        [15 20]>

    >>> print(jnp.multiply(5, v))
    <CartesianVel2D (x[m / s], y[m / s])
        [15 20]>

    """
    # Validation
    lhs = eqx.error_if(
        lhs, any(jax.numpy.shape(lhs)), f"must be a scalar, not {type(lhs)}"
    )

    # Scale the components
    return replace(rhs, x=lhs * rhs.x, y=lhs * rhs.y)


@register(jax.lax.mul_p)
def mul_p_va(lhs: ArrayLike, rhs: CartesianAcc2D, /) -> CartesianAcc2D:
    """Scale a cartesian 2D acceleration by a scalar.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import coordinax as cx

    >>> v = cx.vecs.CartesianAcc2D.from_([3, 4], "m/s2")
    >>> jnp.multiply(5, v).x
    Quantity['acceleration'](Array(15, dtype=int32), unit='m / s2')

    >>> (5 * v).x
    Quantity['acceleration'](Array(15, dtype=int32), unit='m / s2')

    """
    # Validation
    lhs = eqx.error_if(
        lhs, any(jax.numpy.shape(lhs)), f"must be a scalar, not {type(lhs)}"
    )

    # Scale the components
    return replace(rhs, x=lhs * rhs.x, y=lhs * rhs.y)


@register(jax.lax.mul_p)
def mul_p_v_polar(lhs: ArrayLike, rhs: PolarPos, /) -> PolarPos:
    """Scale the polar position by a scalar.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx
    >>> import quaxed

    >>> v = cx.vecs.PolarPos(r=u.Quantity(1, "m"), phi=u.Quantity(90, "deg"))
    >>> print(v)
    <PolarPos (r[m], phi[deg])
        [ 1 90]>

    >>> quaxed.numpy.linalg.vector_norm(v, axis=-1)
    BareQuantity(Array(1., dtype=float32), unit='m')

    >>> nv = quaxed.lax.mul(2, v)
    >>> print(nv)
    <PolarPos (r[m], phi[deg])
        [ 2 90]>

    """
    # Validation
    lhs = eqx.error_if(
        lhs, any(jax.numpy.shape(lhs)), f"must be a scalar, not {type(lhs)}"
    )
    # Scale the radial distance
    return replace(rhs, r=lhs * rhs.r)


# ------------------------------------------------


@register(jax.lax.neg_p)
def neg_p_cart2d_pos(obj: CartesianPos2D, /) -> CartesianPos2D:
    """Negate the `coordinax.vecs.CartesianPos2D`.

    Examples
    --------
    >>> import coordinax as cx
    >>> q = cx.vecs.CartesianPos2D.from_([1, 2], "km")
    >>> (-q).x
    Quantity['length'](Array(-1, dtype=int32), unit='km')

    """
    return jax.tree.map(qlax.neg, obj)


# ------------------------------------------------


@register(jax.lax.sub_p)
def sub_p_cart2d_pos2d(lhs: CartesianPos2D, rhs: AbstractPos, /) -> CartesianPos2D:
    """Subtract two vectors.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx
    >>> cart = cx.vecs.CartesianPos2D.from_([1, 2], "km")
    >>> polr = cx.vecs.PolarPos(r=u.Quantity(3, "km"), phi=u.Quantity(90, "deg"))

    >>> print(cart - polr)
    <CartesianPos2D (x[km], y[km])
        [ 1. -1.]>

    """
    cart = rhs.vconvert(CartesianPos2D)
    return jax.tree.map(jnp.subtract, lhs, cart)



"""Intra-ecosystem Compatibility."""

__all__: list[str] = []


from typing import TypeVar

from jax import Array
from jaxtyping import Shaped
from plum import conversion_method
from zeroth import zeroth

import quaxed.numpy as jnp
from dataclassish import field_values

from .api import cartesian_vector_type
from coordinax._src.vectors.base import AbstractVector
from coordinax._src.vectors.utils import full_shaped

T = TypeVar("T")


@conversion_method(type_from=AbstractVector, type_to=Array)  # type: ignore[arg-type,type-abstract]
def vec_diff_to_array(obj: AbstractVector, /) -> Shaped[Array, "*batch N"]:
    """`coordinax.AbstractVector` -> `Array`.

    Examples
    --------
    >>> import coordinax as cx
    >>> from plum import convert
    >>> import unxt as u

    >>> pos = cx.vecs.CartesianPos1D.from_([1.0], "km")
    >>> convert(pos, Array)
    Array([1.], dtype=float32)

    >>> vel = cx.vecs.CartesianVel1D.from_([1.0], "km/s")
    >>> convert(vel, Array)
    Array([1.], dtype=float32)

    >>> acc = cx.vecs.CartesianAcc1D.from_([1.0], "km/s2")
    >>> convert(acc, Array)
    Array([1.], dtype=float32)

    >>> pos = cx.vecs.RadialPos.from_([1.0], "km")
    >>> convert(pos, Array)
    Array([1.], dtype=float32)

    >>> pos = cx.vecs.CartesianPos2D.from_([1, 2], "km")
    >>> convert(pos, Array)
    Array([1, 2], dtype=int32)

    >>> vel = cx.vecs.CartesianVel2D.from_([1, 2], "km/s")
    >>> convert(vel, Array)
    Array([1, 2], dtype=int32)

    >>> acc = cx.vecs.CartesianAcc2D.from_([1, 2], "km/s2")
    >>> convert(acc, Array)
    Array([1, 2], dtype=int32)

    >>> pos = cx.vecs.PolarPos(u.Quantity(1, "km"), u.Quantity(0, "deg"))
    >>> convert(pos, Array)
    Array([1., 0.], dtype=float32, ...)

    >>> pos = cx.CartesianPos3D.from_([1.0, 2.0, 3.0], "km")
    >>> convert(pos, Array)
    Array([1., 2., 3.], dtype=float32)

    >>> vel = cx.CartesianVel3D.from_([1.0, 2.0, 3.0], "km/s")
    >>> convert(vel, Array)
    Array([1., 2., 3.], dtype=float32)

    >>> acc = cx.vecs.CartesianAcc3D.from_([1.0, 2.0, 3.0], "km/s2")
    >>> convert(acc, Array)
    Array([1., 2., 3.], dtype=float32)

    >>> pos = cx.SphericalPos(u.Quantity(1.0, "km"), u.Quantity(0, "deg"), u.Quantity(0, "deg"))
    >>> convert(pos, Array)
    Array([0., 0., 1.], dtype=float32, ...)

    >>> pos = cx.vecs.CylindricalPos(u.Quantity(1, "km"), u.Quantity(0, "deg"), u.Quantity(0, "km"))
    >>> convert(pos, Array)
    Array([1., 0., 0.], dtype=float32, ...)

    >>> pos = cx.vecs.CartesianPosND.from_([1.0, 2.0, 3.0], "km")
    >>> convert(pos, Array)
    Array([[1.],
           [2.],
           [3.]], dtype=float32)

    >>> vel = cx.vecs.CartesianVelND.from_([1.0, 2.0, 3.0], "km/s")
    >>> convert(vel, Array)
    Array([[1.],
           [2.],
           [3.]], dtype=float32)

    """  # noqa: E501
    cart = obj.vconvert(cartesian_vector_type(obj))  # convert vector to Cartesian
    cart = full_shaped(cart)  # ensure full shape
    comp_qs = field_values(cart)
    unit = zeroth(comp_qs).unit
    comp_arrs = tuple(x.ustrip(unit) for x in field_values(cart))
    return jnp.stack(comp_arrs, axis=-1)



"""Built-in vector classes."""

__all__: list[str] = []


from jaxtyping import Shaped
from plum import conversion_method

from unxt.quantity import AbstractQuantity

from .cartesian import CartesianAccND, CartesianPosND, CartesianVelND


@conversion_method(CartesianAccND, AbstractQuantity)  # type: ignore[arg-type]
@conversion_method(CartesianVelND, AbstractQuantity)  # type: ignore[arg-type]
@conversion_method(CartesianPosND, AbstractQuantity)  # type: ignore[arg-type]
def vec_to_q(
    obj: CartesianPosND | CartesianVelND | CartesianAccND, /
) -> Shaped[AbstractQuantity, "*batch N"]:
    """`coordinax.AbstractPos3D` -> `unxt.Quantity`.

    Examples
    --------
    >>> from plum import convert
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.vecs.CartesianPosND(u.Quantity([1, 2, 3, 4, 5], unit="km"))
    >>> convert(vec, u.Quantity)
    Quantity['length'](Array([1, 2, 3, 4, 5], dtype=int32), unit='km')

    >>> vec = cx.vecs.CartesianVelND(u.Quantity([1, 2, 3, 4, 5], unit="km/s"))
    >>> convert(vec, u.Quantity)
    Quantity['speed'](Array([1, 2, 3, 4, 5], dtype=int32), unit='km / s')

    >>> vec = cx.vecs.CartesianAccND(u.Quantity([1, 2, 3, 4, 5], unit="km/s2"))
    >>> convert(vec, u.Quantity)
    Quantity['acceleration'](Array([1, 2, 3, 4, 5], dtype=int32), unit='km / s2')

    """
    return obj.q



"""Built-in 4-vector classes."""

__all__ = ["FourVector"]

import functools as ft
from dataclasses import KW_ONLY
from typing import Any, cast, final
from typing_extensions import override

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
from jaxtyping import Shaped

import quaxed.numpy as jnp
import unxt as u
from dataclassish import field_values
from dataclassish.converters import Unless

import coordinax._src.custom_types as ct
from .base import AbstractPos4D
from coordinax._src.distances import BBtLength
from coordinax._src.vectors import dims
from coordinax._src.vectors.base import AttrFilter, VectorAttribute
from coordinax._src.vectors.d3 import AbstractPos3D, CartesianPos3D

##############################################################################
# Position


@final
class FourVector(AbstractPos4D):
    """3+1 vector representation.

    The 3+1 vector representation is a 4-vector with 3 spatial coordinates and 1
    time coordinate.

    Parameters
    ----------
    t, q
        Time and spatial coordinates.
    c
        Speed of light, by default ``Quantity(299_792.458, "km/s")``.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    Create a 3+1 vector with a time and 3 spatial coordinates:

    >>> w = cx.FourVector(t=u.Quantity(1, "s"), q=u.Quantity([1, 2, 3], "m"))
    >>> print(w)
    <FourVector (t[s], q=(x[m], y[m], z[m]))
        [1 1 2 3]>

    Note that we used a shortcut to create the 3D vector by passing a ``(*batch,
    3)`` array to the `q` argument. This assumes that `q` is a
    `coordinax.CartesianPos3D` and uses the
    :meth:`coordinax.CartesianPos3D.from_` method to create the 3D vector.

    We can also create a 3D vector explicitly:

    >>> q = cx.SphericalPos(theta=u.Quantity(1, "deg"), phi=u.Quantity(2, "deg"),
    ...                     r=u.Quantity(3, "m"))
    >>> w = cx.FourVector(t=u.Quantity(1, "s"), q=q)
    >>> print(w)
    <FourVector (t[s], q=(r[m], theta[deg], phi[deg]))
        [1 3 1 2]>

    """

    t: ct.BBtTime | ct.ScalarTime = eqx.field(converter=u.Quantity["time"].from_)
    """Time coordinate."""

    q: AbstractPos3D = eqx.field(converter=Unless(AbstractPos3D, CartesianPos3D.from_))
    """Spatial coordinates."""

    _: KW_ONLY
    c: Shaped[u.Quantity["speed"], ""] = eqx.field(
        default=VectorAttribute(default=u.Quantity(299_792.458, "km/s")), repr=False
    )
    """Speed of light, by default ``Quantity(299_792.458, "km/s")``."""

    def __check_init__(self) -> None:
        """Check that the initialization is valid."""
        # Check the shapes are the same, allowing for broadcasting of the time.
        shape = jnp.broadcast_shapes(self.t.shape, self.q.shape)
        if shape != self.q.shape:
            msg = "t and q must be broadcastable to the same shape."
            raise ValueError(msg)

    # TODO: merge with `AvalMixin` and generalize!
    def aval(self) -> jax.core.ShapedArray:
        avals = (self.t.aval(), self.q.aval())
        shape = (*jnp.broadcast_shapes(avals[0].shape, avals[1].shape[:-1]), 4)
        dtype = jnp.result_type(*map(jnp.dtype, avals))
        return jax.core.ShapedArray(shape, dtype)

    # ===============================================================

    def __getattr__(self, name: str) -> Any:
        """Get the attribute from the 3-vector.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> w = cx.FourVector(t=u.Quantity(1, "s"), q=u.Quantity([1, 2, 3], "m"))
        >>> w.x
        Quantity['length'](Array(1, dtype=int32), unit='m')

        """
        return getattr(self.q, name)

    # -------------------------------------------

    @ft.partial(eqx.filter_jit, inline=True)
    def _norm2(self) -> Shaped[u.Quantity["area"], "*#batch"]:
        r"""Return the squared vector norm :math:`(ct)^2 - (x^2 + y^2 + z^2)`.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> w = cx.FourVector(t=u.Quantity(1, "s"), q=u.Quantity([1, 2, 3], "m"))
        >>> w._norm2()
        Quantity['area'](Array(8.987552e+10, dtype=float32), unit='km2')

        """
        return (self.c * self.t) ** 2 - (self.q.norm() ** 2)  # type: ignore[misc,operator]

    @override
    @ft.partial(eqx.filter_jit, inline=True)
    def norm(self) -> BBtLength:
        r"""Return the vector norm :math:`\sqrt{(ct)^2 - (x^2 + y^2 + z^2)}`.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> w = cx.FourVector(t=u.Quantity(1, "s"), q=u.Quantity([1, 2, 3], "m"))
        >>> w.norm()
        Quantity['length'](Array(299792.47+0.j, dtype=complex64), unit='km')

        """
        norm2 = jnp.asarray(self._norm2(), dtype=complex)  # type: ignore[misc]
        return jnp.sqrt(norm2)

    @override
    @property
    def dimensions(self) -> dict[str, u.dims.AbstractDimension]:  # type: ignore[override]
        """Vector physical dimensions.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> cx.FourVector.dimensions
        <property object at ...>

        >>> w = cx.FourVector(t=u.Quantity(1, "s"), q=u.Quantity([1, 2, 3], "m"))
        >>> w.dimensions
        {'t': PhysicalType('time'),
         'q': {'x': PhysicalType('length'), 'y': PhysicalType('length'),
               'z': PhysicalType('length')}}

        """
        return {"t": dims.T, "q": self.q.dimensions}

    # -------------------------------------------
    # misc

    def __str__(self) -> str:
        r"""Return a string representation of the spacetime vector.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx
        >>> w = cx.FourVector(t=u.Quantity(0.5, "s"), q=u.Quantity([1, 2, 3], "m"))
        >>> print(w)
        <FourVector (t[s], q=(x[m], y[m], z[m]))
            [0.5 1.  2.  3. ]>

        """
        cls_name = type(self).__name__
        qcomps = ", ".join(f"{c}[{self.q.units[c]}]" for c in self.q.components)
        comps = f"t[{self.units['t']}], q=({qcomps})"
        vs = np.array2string(  # type: ignore[call-overload]  # TODO: use other method
            jnp.stack(
                tuple(
                    u.ustrip(v)
                    for v in cast(
                        list[u.AbstractQuantity],
                        jnp.broadcast_arrays(self.t, *field_values(AttrFilter, self.q)),
                    )
                ),
                axis=-1,
            ),
            precision=3,
            prefix="    ",
        )
        return f"<{cls_name} ({comps})\n    {vs}>"



"""Copyright (c) 2023 coordinax maintainers. All rights reserved."""

__all__: list[str] = []

import functools as ft
from inspect import isclass
from typing import cast

import equinox as eqx
from jaxtyping import Array, Shaped
from plum import dispatch

import quaxed.numpy as jnp
import unxt as u

from . import api
from coordinax._src.vectors.base import AbstractVector


@dispatch
@ft.partial(eqx.filter_jit, inline=True)
def normalize_vector(x: Shaped[Array, "*batch N"], /) -> Shaped[Array, "*batch N"]:
    """Return the unit vector.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import coordinax as cx

    >>> x = jnp.asarray([2, 0])
    >>> cx.vecs.normalize_vector(x)
    Array([1., 0.], dtype=float32)

    >>> x = jnp.asarray([0, 2])
    >>> cx.vecs.normalize_vector(x)
    Array([0., 1.], dtype=float32)

    """
    return x / jnp.linalg.vector_norm(x, axis=-1, keepdims=True)


@dispatch
@ft.partial(eqx.filter_jit, inline=True)
def normalize_vector(
    x: Shaped[u.AbstractQuantity, "*batch N"], /
) -> Shaped[u.AbstractQuantity, "*batch N"]:
    """Return the unit vector.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = u.Quantity(jnp.asarray([2, 0]), "km")
    >>> cx.vecs.normalize_vector(x)
    Quantity['dimensionless'](Array([1., 0.], dtype=float32), unit='')

    >>> x = u.Quantity(jnp.asarray([0, 2]), "s")
    >>> cx.vecs.normalize_vector(x)
    Quantity['dimensionless'](Array([0., 1.], dtype=float32), unit='')

    """
    return x / jnp.linalg.vector_norm(x, axis=-1, keepdims=True)


# ===========================================================================


@dispatch
def time_nth_derivative_vector_type(
    obj: type[AbstractVector] | AbstractVector, /, *, n: int
) -> type[AbstractVector]:
    out = cast(type[AbstractVector], obj if isclass(obj) else type(obj))
    if n == 0:
        pass
    elif n < 0:
        for _ in range(-n):
            out = api.time_antiderivative_vector_type(out)
    else:
        for _ in range(n):
            out = api.time_derivative_vector_type(out)

    return out



"""Sequence of Operators."""

__all__ = ["Pipe", "convert_to_pipe_operators"]

from dataclasses import replace
from typing import Any, final

import equinox as eqx
import wadler_lindig as wl
from plum import dispatch

from .base import AbstractOperator
from .composite import AbstractCompositeOperator


@dispatch.abstract
def convert_to_pipe_operators(inp: Any, /) -> tuple[AbstractOperator, ...]:
    raise NotImplementedError  # pragma: no cover


@final
class Pipe(AbstractCompositeOperator):
    r"""Operator Pipe.

    Piping refers to a process in which the output of one operation is directly
    passed as the input to another. This is a composite operator that represents
    a sequence of operations to be applied in order.

    `Pipe` operators can be created using the 'pipe' syntax `op1 | op2`. A Pipe
    operator created as ``FG = F | G``, when evaluated, is equivalent to
    evaluating $g \circ f = g(f(x))$. Note the order of the operators!

    :::{note}

    The `|` operator works differently from the functional composition operator
    $\circ$, which is sadly not supported in Python. The `|` operator is like
    the Unix Shell pipe operator, where output is passed left-to-right. This
    order can be seen in the indexing of the operators in the `Pipe` object.

    :::

    Parameters
    ----------
    operators
        The sequence of operators to apply.

    Examples
    --------
    >>> import coordinax as cx

    >>> shift = cx.ops.GalileanSpatialTranslation.from_([1, 2, 3], "km")
    >>> boost = cx.ops.VelocityBoost.from_([1, 2, 3], "km/s")
    >>> pipe = cx.ops.Pipe((shift, boost))
    >>> pipe
    Pipe(( GalileanSpatialTranslation(...), VelocityBoost(...) ))

    A pipe can also be constructed by ``|``:

    >>> pipe2 = shift | boost
    >>> pipe2
    Pipe(( GalileanSpatialTranslation(...), VelocityBoost(...) ))

    The pipe can be simplified. For this example, we add an identity operator to
    the sequence and simplify, which will remove the identity operator.

    >>> pipe3 = pipe2 | cx.ops.Identity()
    >>> pipe3
    Pipe((
        GalileanSpatialTranslation(...), VelocityBoost(...), Identity()
    ))

    >>> cx.ops.simplify_op(pipe3)
    Pipe(( GalileanSpatialTranslation(...), VelocityBoost(...) ))

    Now let's call the operator on a position:

    >>> pos = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> print(pipe(pos))
    <CartesianPos3D (x[km], y[km], z[km])
        [2 4 6]>

    The pipe will also work on a position + velocity:

    >>> vel = cx.CartesianVel3D.from_([4, 5, 6], "km/s")
    >>> print(*pipe(pos, vel), sep="\n")
    <CartesianPos3D (x[km], y[km], z[km])
        [2 4 6]>
    <CartesianVel3D (x[km / s], y[km / s], z[km / s])
        [5. 7. 9.]>

    """

    operators: tuple[AbstractOperator, ...] = eqx.field(
        converter=convert_to_pipe_operators
    )

    # ---------------------------------------------------------------

    def __or__(self, other: AbstractOperator) -> "Pipe":
        """Compose with another operator.

        Examples
        --------
        >>> import coordinax as cx

        >>> shift = cx.ops.GalileanSpatialTranslation.from_([1, 2, 3], "km")
        >>> boost = cx.ops.GalileanBoost.from_([1, 2, 3], "km/s")
        >>> pipe = cx.ops.Pipe((shift, boost))

        >>> pipe | pipe
        Pipe(( GalileanSpatialTranslation(...), GalileanBoost(...),
               GalileanSpatialTranslation(...), GalileanBoost(...) ))

        """
        # Concatenate sequences
        if isinstance(other, type(self)):
            return replace(self, operators=self.operators + other.operators)
        # Append single operators
        return replace(self, operators=(*self, other))

    def __ror__(self, other: AbstractOperator) -> "Pipe":
        """Compose with another operator."""
        # Append single operators
        return replace(self, operators=(other, *self))

    def __pdoc__(self, **kwargs: Any) -> wl.AbstractDoc:
        """Return the Wadler-Lindig representation."""
        docs = [wl.pdoc(op, **kwargs) for op in self.operators]
        begin = wl.TextDoc("((" if len(docs) > 1 else "(")
        end = wl.TextDoc("))" if len(docs) > 1 else ")")
        return wl.bracketed(
            begin=wl.TextDoc(f"{self.__class__.__name__}") + begin,
            docs=docs,
            sep=wl.comma,
            end=end,
            indent=kwargs.get("indent", 4),
        )


# ==============================================================
# Constructor


@dispatch
def convert_to_pipe_operators(
    inp: tuple[AbstractOperator, ...] | list[AbstractOperator],
) -> tuple[AbstractOperator, ...]:
    """Convert to a tuple of operators.

    Examples
    --------
    >>> import coordinax as cx

    >>> op1 = cx.ops.GalileanRotation([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    >>> op2 = cx.ops.Identity()
    >>> convert_to_pipe_operators((op1, op2))
    (GalileanRotation(rotation=i32[3,3]), Identity())

    """
    return tuple(inp)


@dispatch
def convert_to_pipe_operators(inp: AbstractOperator) -> tuple[AbstractOperator, ...]:
    """Convert to a tuple of operators.

    Examples
    --------
    >>> import coordinax as cx

    >>> op1 = cx.ops.GalileanRotation([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    >>> convert_to_pipe_operators(op1)
    (GalileanRotation(rotation=i32[3,3]),)

    """
    return (inp,)


@dispatch
def convert_to_pipe_operators(inp: Pipe) -> tuple[AbstractOperator, ...]:
    """Convert to a tuple of operators.

    Examples
    --------
    >>> import coordinax as cx

    >>> op1 = cx.ops.Identity()
    >>> op2 = cx.ops.Identity()
    >>> pipe = cx.ops.Pipe((op1, op2))
    >>> convert_to_pipe_operators(pipe)
    (Identity(), Identity())

    """
    return inp.operators
